
import UIKit
import CloudKit
import CoreData

class ApplicationData {
    var context: NSManagedObjectContext!
    var database: CKDatabase!
    
    init() {
        let app = UIApplication.shared
        let appDelegate = app.delegate as! AppDelegate
        context = appDelegate.context
        
        let container = CKContainer.default()
        database = container.privateCloudDatabase
    }
    func checkUpdates(finishClosure: @escaping (UIBackgroundFetchResult) -> Void) {
        configureDatabase(executeClosure: {
            self.downloadUpdates(finishClosure: finishClosure)
        })
    }
    func configureDatabase(executeClosure: @escaping () -> Void) {
        let userSettings = UserDefaults.standard
        if !userSettings.bool(forKey: "subscriptionSaved") {
            let newSubscription = CKDatabaseSubscription(subscriptionID: "updatesDatabase")
            let info = CKSubscription.NotificationInfo()
            info.shouldSendContentAvailable = true
            newSubscription.notificationInfo = info
            
            database.save(newSubscription, completionHandler: { (subscription, error) in
                if error != nil {
                    print("Error Creating Subscription")
                } else {
                    userSettings.set(true, forKey: "subscriptionSaved")
                }
            })
        }
        if !userSettings.bool(forKey: "zoneCreated") {
            let newZone = CKRecordZone(zoneName: "listPlaces")
            database.save(newZone, completionHandler: { (zone, error) in
                if error != nil {
                    print("Error Creating Zone")
                } else {
                    userSettings.set(true, forKey: "zoneCreated")
                    executeClosure()
                }
            })
        } else {
            executeClosure()
        }
    }
    func downloadUpdates(finishClosure: @escaping (UIBackgroundFetchResult) -> Void) {
        var listRecordsUpdated: [CKRecord] = []
        var listRecordsDeleted: [String: String] = [:]
        var changeToken: CKServerChangeToken!
        var changeZoneToken: CKServerChangeToken!
        
        let userSettings = UserDefaults.standard
        if let data = userSettings.value(forKey: "changeToken") as? Data {
            if let token = try? NSKeyedUnarchiver.unarchivedObject(ofClass: CKServerChangeToken.self, from: data) {
                changeToken = token
            }
        }
        if let data = userSettings.value(forKey: "changeZoneToken") as? Data {
            if let token = try? NSKeyedUnarchiver.unarchivedObject(ofClass: CKServerChangeToken.self, from: data) {
                changeZoneToken = token
            }
        }
        var zonesIDs: [CKRecordZone.ID] = []
        let operation = CKFetchDatabaseChangesOperation(previousServerChangeToken: changeToken)
        operation.recordZoneWithIDChangedBlock = { (zoneID) in
            zonesIDs.append(zoneID)
        }
        operation.changeTokenUpdatedBlock = { (token) in
            changeToken = token
        }
        operation.fetchDatabaseChangesCompletionBlock = { (token, more, error) in
            if error != nil {
                finishClosure(UIBackgroundFetchResult.failed)
            } else if !zonesIDs.isEmpty {
                changeToken = token
                
                let configuration = CKFetchRecordZoneChangesOperation.ZoneConfiguration()
                configuration.previousServerChangeToken = changeZoneToken
                let fetchOperation = CKFetchRecordZoneChangesOperation(recordZoneIDs: zonesIDs, configurationsByRecordZoneID: [zonesIDs[0]: configuration])
                fetchOperation.recordChangedBlock = { (record) in
                    listRecordsUpdated.append(record)
                }
                fetchOperation.recordWithIDWasDeletedBlock = { (recordID, recordType) in
                    listRecordsDeleted[recordID.recordName] = recordType
                }
                fetchOperation.recordZoneChangeTokensUpdatedBlock = { (zoneID, token, data) in
                    changeZoneToken = token
                }
                fetchOperation.recordZoneFetchCompletionBlock = { (zoneID, token, data, more, error) in
                    if error != nil {
                        print("Error")
                    } else {
                        changeZoneToken = token
                        
                        self.updateLocalRecords(listRecordsUpdated: listRecordsUpdated)
                        self.deleteLocalRecords(listRecordsDeleted: listRecordsDeleted)
                        listRecordsUpdated.removeAll()
                        listRecordsDeleted.removeAll()
                    }
                }
                fetchOperation.fetchRecordZoneChangesCompletionBlock = { (error) in
                    if error != nil {
                        print("Error")
                        finishClosure(UIBackgroundFetchResult.failed)
                    } else {
                        if changeToken != nil {
                            if let data = try? NSKeyedArchiver.archivedData(withRootObject: changeToken, requiringSecureCoding: false) {
                                userSettings.set(data, forKey: "changeToken")
                            }
                        }
                        if changeZoneToken != nil {
                            if let data = try? NSKeyedArchiver.archivedData(withRootObject: changeZoneToken, requiringSecureCoding: false) {
                                userSettings.set(data, forKey: "changeZoneToken")
                            }
                        }
                        self.updateLocalReferences()
                        finishClosure(UIBackgroundFetchResult.newData)
                    }
                }
                self.database.add(fetchOperation)
            } else {
                finishClosure(UIBackgroundFetchResult.noData)
            }
        }
        database.add(operation)
    }
    func updateLocalRecords(listRecordsUpdated: [CKRecord]) {
        for record in listRecordsUpdated {
            let coder = NSKeyedArchiver(requiringSecureCoding: true)
            record.encodeSystemFields(with: coder)
            let recordMetadata = coder.encodedData
            
            let recordType = record.recordType
            let recordName = record.recordID.recordName
            
            if recordType == "Countries" {
                let request: NSFetchRequest<Countries> = Countries.fetchRequest()
                request.predicate = NSPredicate(format: "ckrecordname = %@", recordName)
                do {
                    var country: Countries!
                    let result = try context.fetch(request)
                    if result.isEmpty {
                        country = Countries(context: context)
                        country.ckrecordname = recordName
                    } else {
                        country = result[0]
                    }
                    country.ckmetadata = recordMetadata
                    country.ckupload = false
                    country.name = record["name"] as? String
                } catch {
                    print("Error Fetching")
                }
            } else if recordType == "Cities" {
                let request: NSFetchRequest<Cities> = Cities.fetchRequest()
                request.predicate = NSPredicate(format: "ckrecordname = %@", recordName)
                do {
                    var city: Cities!
                    let result = try context.fetch(request)
                    if result.isEmpty {
                        city = Cities(context: context)
                        city.ckrecordname = recordName
                    } else {
                        city = result[0]
                    }
                    city.ckmetadata = recordMetadata
                    city.ckupload = false
                    city.ckpicture = false
                    city.name = record["name"] as? String
                    if let reference = record["country"] as? CKRecord.Reference {
                        city.ckreference = reference.recordID.recordName
                    }
                    if let asset = record["picture"] as? CKAsset {
                        let picture = UIImage(contentsOfFile: asset.fileURL.path)
                        city.picture = picture?.pngData()
                    }
                } catch {
                    print("Error Fetching")
                }
            }
        }
        if context.hasChanges {
            do {
                try context.save()
            } catch {
                print("Error Saving Context")
            }
        }
    }
    func deleteLocalRecords(listRecordsDeleted: [String: String]) {
        for (recordName, recordType) in listRecordsDeleted {
            if recordType == "Countries" {
                let request: NSFetchRequest<Countries> = Countries.fetchRequest()
                request.predicate = NSPredicate(format: "ckrecordname = %@", recordName)
                do {
                    let result = try context.fetch(request)
                    if !result.isEmpty {
                        let country = result[0]
                        context.delete(country)
                    }
                } catch {
                    print("Error Fetching")
                }
            } else if recordType == "Cities" {
                let request: NSFetchRequest<Cities> = Cities.fetchRequest()
                request.predicate = NSPredicate(format: "ckrecordname = %@", recordName)
                do {
                    let result = try context.fetch(request)
                    if !result.isEmpty {
                        let city = result[0]
                        context.delete(city)
                    }
                } catch {
                    print("Error Fetching")
                }
            }
        }
        if context.hasChanges {
            do {
                try context.save()
            } catch {
                print("Error Saving Context")
            }
        }
    }
    func updateLocalReferences() {
        let requestCities: NSFetchRequest<Cities> = Cities.fetchRequest()
        requestCities.predicate = NSPredicate(format: "ckreference != nil")
        do {
            let listCities = try context.fetch(requestCities)
            for city in listCities {
                let requestCountries: NSFetchRequest<Countries> = Countries.fetchRequest()
                requestCountries.predicate = NSPredicate(format: "ckrecordname = %@", city.ckreference!)
                do {
                    let listCountries = try context.fetch(requestCountries)
                    if !listCountries.isEmpty {
                        city.country = listCountries[0]
                        city.ckreference = nil
                    }
                } catch {
                    print("Error Fetching")
                }
            }
        } catch {
            print("Error Fetching")
        }
        if context.hasChanges {
            do {
                try context.save()
            } catch {
                print("Error Saving Context")
            }
        }
    }
    func uploadRecords() {
        let listRecordsToUpload = getRecordsToUpload()
        if !listRecordsToUpload.isEmpty {
            configureDatabase(executeClosure: {
                let operation = CKModifyRecordsOperation(recordsToSave: listRecordsToUpload, recordIDsToDelete: nil)
                operation.modifyRecordsCompletionBlock = { (records, recordsID, error) in
                    if error != nil {
                        print("Error uploading records")
                    } else {
                        self.clearCoreDataRecords()
                    }
                }
                self.database.add(operation)
            })
        }
    }
    func getRecordsToUpload() -> [CKRecord] {
        var list: [CKRecord] = []
        let requestCountries: NSFetchRequest<Countries> = Countries.fetchRequest()
        requestCountries.predicate = NSPredicate(format: "ckupload = true")
        do {
            let result = try context.fetch(requestCountries)
            for item in result {
                var recordtemp: CKRecord!
                if let coder = try? NSKeyedUnarchiver(forReadingFrom: item.ckmetadata!) {
                    coder.requiresSecureCoding = true
                    recordtemp = CKRecord(coder: coder)
                    coder.finishDecoding()
                }
                if let record = recordtemp {
                    record.setObject(item.name! as NSString, forKey: "name")
                    list.append(record)
                }
            }
        } catch {
            print("Error Fetching")
        }
        let requestCities: NSFetchRequest<Cities> = Cities.fetchRequest()
        requestCities.predicate = NSPredicate(format: "ckupload = true")
        do {
            let result = try context.fetch(requestCities)
            for item in result {
                var recordtemp: CKRecord!
                if let coder = try? NSKeyedUnarchiver(forReadingFrom: item.ckmetadata!) {
                    coder.requiresSecureCoding = true
                    recordtemp = CKRecord(coder: coder)
                    coder.finishDecoding()
                }
                if let record = recordtemp {
                    record.setObject(item.name! as NSString, forKey: "name")
                    
                    if item.ckpicture && item.picture != nil {
                        var url = URL(fileURLWithPath: NSTemporaryDirectory())
                        url = url.appendingPathComponent("tempfile.png")
                        do {
                            let pngImage = item.picture!
                            try pngImage.write(to: url, options: .atomic)
                            let asset = CKAsset(fileURL: url)
                            record.setObject(asset, forKey: "picture")
                        } catch {
                            print("Image not stored")
                        }
                    }
                    if let country = item.country {
                        let zone = CKRecordZone(zoneName: "listPlaces")
                        let id = CKRecord.ID(recordName: country.ckrecordname!, zoneID: zone.zoneID)
                        let reference = CKRecord.Reference(recordID: id, action: .deleteSelf)
                        record.setObject(reference, forKey: "country")
                    }
                    list.append(record)
                }
            }
        } catch {
            print("Error Fetching")
        }
        return list
    }
    func clearCoreDataRecords() {
        let requestCountries: NSFetchRequest<Countries> = Countries.fetchRequest()
        requestCountries.predicate = NSPredicate(format: "ckupload = true")
        do {
            let result = try context.fetch(requestCountries)
            for item in result {
                item.ckupload = false
            }
        } catch {
            print("Error Fetching")
        }
        let requestCities: NSFetchRequest<Cities> = Cities.fetchRequest()
        requestCities.predicate = NSPredicate(format: "ckupload = true")
        do {
            let result = try context.fetch(requestCities)
            for item in result {
                item.ckupload = false
                item.ckpicture = false
            }
        } catch {
            print("Error Fetching")
        }
        if context.hasChanges {
            do {
                try context.save()
            } catch {
                print("Error Saving Context")
            }
        }
    }
    func removeRecords() {
        var listRecordsToDelete: [CKRecord.ID] = []
        
        let request: NSFetchRequest<CKDelete> = CKDelete.fetchRequest()
        do {
            let result = try context.fetch(request)
            for item in result {
                let zone = CKRecordZone(zoneName: item.zoneName!)
                let id = CKRecord.ID(recordName: item.recordName!, zoneID: zone.zoneID)
                listRecordsToDelete.append(id)
            }
        } catch {
            print("Error fetching record IDs")
        }
        if !listRecordsToDelete.isEmpty {
            let operation = CKModifyRecordsOperation(recordsToSave: nil, recordIDsToDelete: listRecordsToDelete)
            operation.modifyRecordsCompletionBlock = { (records, recordsID, error) in
                if error != nil {
                    print("Error deleting records")
                } else {
                    let request: NSFetchRequest<CKDelete> = CKDelete.fetchRequest()
                    do {
                        let result = try self.context.fetch(request)
                        for item in result {
                            self.context.delete(item)
                        }
                        try self.context.save()
                    } catch {
                        print("Error deleting")
                    }
                }
            }
            database.add(operation)
        }
    }
}
var AppData = ApplicationData()
